# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from thriftrw.wire import TType
from thriftrw.wire.value import StructValue, FieldValue
from thriftrw.compile.exceptions import ThriftCompilerError

from . import check
from . import common
from .base import TypeSpec
from .const import const_value_or_ref
from .spec_mapper import type_spec_or_ref

__all__ = ['StructTypeSpec', 'FieldSpec']


class StructTypeSpec(TypeSpec):
    """A struct is a collection of named fields.

    The surface for struct types is a class with the following:

    .. py:attribute:: type_spec

        :py:class:`StructTypeSpec` for the type.

    .. py:method:: __init__(self, *args, **kwargs)

        Accepts all fields of the struct as arguments. Required arguments are
        placed first and may be specified as positional arguments (in the same
        order as they appear in the IDL). Optional arguments and required
        arguments come next. A TypeError will be raised if the required
        arguments for the struct are not filled with non-None values.

    And obvious definitions of ``__str__`` and ``__eq__``.

    Given the definition,::

        struct User {
            1: required string name
            2: optional string email
            3: required bool isActive = true
        }

    A class roughly equivalent to the following is generated,

    .. code-block:: python

        class User(object):

            __slots__ = ('name', 'email', 'isActive')

            type_spec = # ...

            def __init__(name, email=None, isActive=True):
                # ...
    """

    __slots__ = (
        'name', 'fields', 'linked', 'surface', 'base_cls',
    )

    ttype_code = TType.STRUCT

    def __init__(self, name, fields, base_cls=None):
        """Initialize a new StructTypeSpec.

        :param str name:
            Name of the Struct.
        :param fields:
            Collection of :py:class:`FieldSpec` objects.
        :param base_cls:
            Base class to use for generates classes. Defaults to ``object``.
        """

        #: Name of the struct.
        self.name = name

        #: Collection of :py:class:`FieldSpec` objects.
        self.fields = fields
        self.linked = False
        self.surface = None
        self.base_cls = base_cls or object

    def link(self, scope):
        if not self.linked:
            self.linked = True
            self.fields = [field.link(scope) for field in self.fields]
            self.surface = struct_cls(self, scope)
        return self

    @classmethod
    def compile(cls, struct, require_requiredness=True):
        fields = []
        ids = set()
        names = set()
        for field in struct.fields:
            if field.name in names:
                raise ThriftCompilerError(
                    'Field "%s" of struct "%s" on line %d has duplicates.' %
                    (field.name, struct.name, field.lineno)
                )
            names.add(field.name)

            if field.id in ids:
                raise ThriftCompilerError(
                    'Field ID "%d" of struct "%s" on line %d '
                    'has already been used.' % (
                        field.id, struct.name, field.lineno
                    )
                )
            ids.add(field.id)
            fields.append(FieldSpec.compile(
                field=field,
                struct_name=struct.name,
                require_requiredness=require_requiredness,
            ))
        return cls(struct.name, fields)

    def to_wire(self, struct):
        check.instanceof_surface(self, struct)
        fields = []

        for field in self.fields:
            value = getattr(struct, field.name)
            if value is None:
                if field.required:
                    raise TypeError(
                        'Field "%s" of "%s" is required. It cannot be None.'
                        % (field.name, self.name)
                    )
                else:
                    continue
            fields.append(field.to_wire(value))

        return StructValue(fields)

    def from_wire(self, wire_value):
        check.type_code_matches(self, wire_value)
        kwargs = {}
        for field in self.fields:
            field_value = wire_value.get(field.id, field.ttype_code)
            if field_value is None:
                continue
            kwargs[field.name] = field.from_wire(field_value)

        # TODO For the case where cls fails to instantiate because a required
        # positional argument is missing, we know that the request was
        # invalid.
        return self.surface(**kwargs)

    def __str__(self):
        return 'StructTypeSpec(name=%r, fields=%r)' % (self.name, self.fields)

    __repr__ = __str__

    def __eq__(self, other):
        return (
            self.name == other.name and
            self.fields == other.fields and
            self.base_cls == other.base_cls and
            self.linked == other.linked
        )


class FieldSpec(object):
    """Specification for a single field on a struct.

    FieldSpecs do not expose anything at the module level.
    """

    __slots__ = ('id', 'name', 'spec', 'required', 'default_value', 'linked')

    def __init__(self, id, name, spec, required, default_value=None):
        #: Field identifier of the field.
        self.id = id

        #: Name of the field.
        self.name = name

        #: TypeSpec for the type of values accepted by the field.
        self.spec = spec

        #: Whether this field is required or not.
        self.required = required

        #: Default value of the field.
        self.default_value = default_value

        self.linked = False

    def link(self, scope):
        if not self.linked:
            self.linked = True
            if self.default_value is not None:
                self.default_value = self.default_value.link(scope).surface
            self.spec = self.spec.link(scope)
        return self

    @property
    def ttype_code(self):
        return self.spec.ttype_code

    @classmethod
    def compile(cls, field, struct_name, require_requiredness=True):
        if field.id is None:
            raise ThriftCompilerError(
                'Field "%s" of "%s" does not have an explicit field ID. '
                'Please specify the numeric ID for the field.'
                % (field.name, struct_name)
            )

        required = field.requiredness
        if required is None:
            if require_requiredness:
                raise ThriftCompilerError(
                    'Field "%s" of "%s" on line %d does not explicitly '
                    'specify requiredness. Please specify whether the field '
                    'is optional or required in the IDL.'
                    % (field.name, struct_name, field.lineno)
                )
            else:
                required = False

        # TODO check field ids are valid signed 16-bit integers

        default_value = None
        if field.default is not None:
            default_value = const_value_or_ref(field.default)

        field_type_spec = type_spec_or_ref(field.field_type)
        return cls(
            id=field.id,
            name=field.name,
            spec=field_type_spec,
            required=required,
            default_value=default_value,
        )

    # While FieldSpec has an interface similar to TypeSpec, it's not an actual
    # TypeSpec.

    def to_wire(self, value):
        assert value is not None
        return FieldValue(
            id=self.id,
            ttype=self.spec.ttype_code,
            value=self.spec.to_wire(value),
        )

    def from_wire(self, wire_value):
        assert wire_value is not None
        return self.spec.from_wire(wire_value.value)

    def __str__(self):
        # Field __str__ must reference spec names instead of the specs to
        # avoid an infinite loop in case the field is self-referential.
        return 'FieldSpec(id=%r, name=%r, spec_name=%r)' % (
            self.id, self.name, self.spec.name
        )

    __repr__ = __str__

    def __eq__(self, other):
        return (
            self.id == other.id and
            self.name == other.name and
            self.spec == other.spec and
            self.required == other.required and
            self.default_value == other.default_value and
            self.linked == other.linked
        )


def struct_init(cls_name, field_names, field_defaults, base_cls):
    """Generate the ``__init__`` method for structs.

    ``field_names`` is a list or tuple of field names for the constructor
    in-order. ``field_defaults`` is a list or tuple of default values for the
    last ``len(field_defaults)`` fields in-order.

    That is, for a signature like, ``f(a, b, c=1, d=2)``, ``field_names`` is
    ``('a', 'b', 'c', 'd')`` and ``field_defaults`` is ``(1, 2)``.

    :param cls_name:
        Name of the class. This is used for error messages.
    :param field_names:
        List or tuple of field names of the struct.
    :param field_defaults:
        List or tuple of default values for the last ``len(field_defaults)``
        fields. For optional fields with no default value, ``field_defaults``
        should have a ``None`` default.
    :returns:
        Implementation of the ``__init__`` method for that struct.
    """
    field_defaults = field_defaults or ()

    # Number of fields on the struct. We can't accept more than this many
    # arguments.
    num_fields = len(field_names)

    # Number of fields with default values.
    num_defaults = len(field_defaults)

    assert num_defaults <= num_fields, 'Too many defaults provided.'

    # The last num_defaults fields have default values, so all other fields
    # must have arguments passed in.
    num_required = num_fields - num_defaults

    def __init__(self, *args, **kwargs):
        base_cls.__init__(self)

        num_args = len(args)
        num_kwargs = len(kwargs)
        num_total = num_args + num_kwargs

        # This argument binding implementation was adapted from
        # inspect.getcallargs() with adjustments for our requirements of
        # non-None fields.

        # TODO pull out argument binding logic into separate function
        # elsewhere.

        if num_total < num_required:
            raise TypeError(
                '%s() takes at least %d arguments (%d given)' % (
                    cls_name, num_required + 1, num_total + 1
                    # + 1 for 'self'
                )
            )

        if num_total > num_fields:
            raise TypeError(
                '%s() takes at most %d arguments (%d given)' % (
                    cls_name, num_fields + 1, num_total + 1
                    # + 1 for 'self'
                )
            )

        # Keeps track of fields yet to be assigned.
        unassigned = set(field_names)

        # Assign positional arguments.
        for name, value in zip(field_names, args):
            if value is not None:
                setattr(self, name, value)
                unassigned.remove(name)

        # Assign named arguments.
        for name in field_names:
            # We go over all field names instead of just the remaining ones so
            # that we can detect if any of the positional arguments had
            # another value specified in kwargs. That's user error and we
            # should blow up in that case.
            if name not in kwargs:
                continue
            if name not in unassigned:
                raise TypeError(
                    '%s() got multiple values for argument "%s"'
                    % (cls_name, name)
                )
            value = kwargs.pop(name)
            if value is not None:
                setattr(self, name, value)
                unassigned.remove(name)

        # Add defaults
        if field_defaults:
            default_values = zip(field_names[-num_defaults:], field_defaults)
            for name, value in default_values:
                if name in unassigned:
                    setattr(self, name, value)
                    unassigned.remove(name)

        if kwargs:
            # Too many kwargs given.
            for name in kwargs:
                raise TypeError(
                    '%s() got an unexpected keyword argument "%s"'
                    % (cls_name, name)
                )

        # If anything was left unassigned, blow up.
        if unassigned:
            raise ValueError(
                'Field(s) %r require non-None values.' % list(unassigned)
            )

    # TODO reasonable docstring
    return __init__


def struct_docstring(cls_name, required_fields, optional_fields):
    """Generates a docstring for generated structs."""
    required_section = ''
    if required_fields:
        required_section = '\n'.join(
            [''] +
            ['\nThe following arguments are required:\n'] +
            ['-   %s' % name for name in required_fields]
        )

    optional_section = ''
    if optional_fields:
        optional_section = '\n'.join(
            [''] +
            ['\nThe following arguments are optional:\n'] +
            ['-   %s (default: %s)' % pair
             for pair in optional_fields.items()]
        )

    param_list = (
        list(required_fields) +
        ['%s=None' % name for name in optional_fields.keys()]
    )

    header = '%s(%s)' % (cls_name, ', '.join(param_list))

    return header + required_section + optional_section


def struct_cls(struct_spec, scope):
    """Generate a class for a struct.

    The generated class has the following attributes:

    :param StructTypeSpec struct_spec:
        Type specification of the struct.
    :param thriftrw.compile.Scope scope:
        The compilation scope
    """
    slots = set()

    required_fields = []
    optional_fields = []
    field_defaults = []

    for field in struct_spec.fields:
        slots.add(field.name)

        default = None
        if field.default_value is not None:
            default = field.default_value

        if field.required and default is None:
            # required fields that have default values are optional as far as
            # the constructor is concerned.
            required_fields.append(field.name)
        else:
            optional_fields.append(field.name)
            field_defaults.append(default)

    field_names = required_fields + optional_fields

    struct_dct = {}
    struct_dct['type_spec'] = struct_spec
    struct_dct['__slots__'] = tuple(slots)
    struct_dct['__init__'] = struct_init(
        struct_spec.name,
        field_names,
        field_defaults,
        struct_spec.base_cls,
    )
    struct_dct['__str__'] = common.fields_str(struct_spec.name, field_names)
    struct_dct['__repr__'] = struct_dct['__str__']
    struct_dct['__eq__'] = common.fields_eq(set(field_names))
    struct_dct['__doc__'] = struct_docstring(
        struct_spec.name,
        required_fields,
        dict(zip(optional_fields, field_defaults)),
    )

    # TODO generate a reasonable docstring for the class too.

    return type(str(struct_spec.name), (struct_spec.base_cls,), struct_dct)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

import abc

from thriftrw.wire import TType


__all__ = ['Protocol']


class Protocol(object):
    """Base class for all protocol implementations."""

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def serialize_value(self, value):
        """Serialize the given ``Value``.

        :param ~thriftrw.wire.Value value:
            Value to serialize.
        :returns:
            Serialized value.
        """

    @abc.abstractmethod
    def deserialize_value(self, typ, s):
        """Parse a ``Value`` of the given type.

        :param ~thriftrw.wire.TType typ:
            Type code of the value to parse.
        :param s:
            Bytes to decode.
        :returns:
            Parsed :py:class:`~thriftrw.wire.Value`.
        :raises thriftrw.protocol.ThriftProtocolError:
            If the object failed to deserialize.
        """

    def dumps(self, obj):
        """Serializes the given Thrift struct.

        :param obj:
            An instance of a class generated by thriftrw representing a
            struct, union, or exceptino.
        :returns:
            Binary representation of the object.
        """
        value = obj.__class__.type_spec.to_wire(obj)
        return self.serialize_value(value)

    def loads(self, obj_cls, s):
        """Deserializes an object from the given blob.

        :param obj_cls:
            A class generated by thriftrw representing a struct, union, or
            exception.
        :param bytes s:
            Binary blob representing the object.
        :returns:
            Deserialized object.
        :raises thriftrw.protocol.ThriftProtocolError:
            If the object failed to deserialize.
        """
        value = self.deserialize_value(TType.STRUCT, s)
        return obj_cls.type_spec.from_wire(value)

    # TODO do we care about Message envelopes?



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from thriftrw.compile.exceptions import ThriftCompilerError

from .spec_mapper import type_spec_or_ref


class ConstValuePrimitive(object):

    __slots__ = ('surface',)

    def __init__(self, value):
        self.surface = value

    def link(self, scope):
        return self


class ContsValueMap(object):

    __slots__ = ('items', 'linked', 'surface')

    def __init__(self, items):
        self.items = items
        self.linked = False
        self.surface = None

    def link(self, scope):
        if not self.linked:
            self.linked = True
            self.surface = {
                k.link(scope).surface: v.link(scope).surface
                for k, v in self.items.items()
            }
        return self


class ConstValueList(object):

    __slots__ = ('values', 'linked', 'surface')

    def __init__(self, values):
        self.values = values
        self.linked = False
        self.surface = None

    def link(self, scope):
        if not self.linked:
            self.linked = True
            self.surface = [v.link(scope).surface for v in self.values]
        return self


class ConstValueReference(object):

    __slots__ = ('name', 'lineno')

    def __init__(self, name, lineno):
        self.name = name
        self.lineno = lineno

    def link(self, scope):
        return scope.resolve_const_spec(self.name, self.lineno)

    def __str__(self):
        return 'ConstValueReference(name=%s, lineno=%d)' % (
            self.name, self.lineno
        )

    __repr__ = __str__


class ConstSpec(object):
    """Spec for a constant value defined in the Thrift file.

    The surface for a ConstSpec is the value defined in the Thrift file.

    The following,::

        const list<i32> foo = [1, 2, 3]

    Roughly translates to,

    .. code-block:: python

        foo = [1, 2, 3]

    In the generated module.
    """

    __slots__ = (
        'name', 'type_spec', 'value_spec', 'linked', 'surface', 'save'
    )

    def __init__(self, name, value_spec, type_spec, save=None):
        """
        :param name:
            Name of the constant.
        :param value_spec:
            spec of the constant value
        :param type_spec:
            TypeSpec for the type of the value. The value will be validated
            against this spec.
        :param save:
            Whether the constant's surface should be exposed in the generated
            module. Defaults to True.
        """

        #: Name of the constant.
        self.name = name

        self.value_spec = value_spec
        self.type_spec = type_spec

        self.linked = False
        self.surface = None

        if save is None:
            save = True
        self.save = save

    @classmethod
    def compile(cls, const):
        # TODO validate that const.name is a valid python identifier.
        type_spec = type_spec_or_ref(const.value_type)
        value_spec = const_value_or_ref(const.value)

        return cls(const.name, value_spec, type_spec)

    def link(self, scope):
        if not self.linked:
            self.linked = True
            self.type_spec = self.type_spec.link(scope)
            self.value_spec = self.value_spec.link(scope)
            value = self.value_spec.surface
            try:
                self.type_spec.to_wire(value)
            except TypeError as e:
                raise ThriftCompilerError(
                    'Value for constant "%s" does not match its type "%s": %s'
                    % (self.name, self.type_spec.name, e)
                )
            except ValueError as e:
                raise ThriftCompilerError(
                    'Value for constant "%s" is not valid: %s'
                    % (self.name, e)
                )
            self.surface = value

        return self


class ConstValueMapper(object):
    """Resolves constant values."""

    __slots__ = ()

    def get(self, const_value):
        return const_value.apply(self)

    def visit_primitive(self, const):
        return ConstValuePrimitive(const.value)

    def visit_list(self, const):
        return ConstValueList([self.get(v) for v in const.values])

    def visit_map(self, const):
        return ContsValueMap({
            self.get(k): self.get(v) for k, v in const.pairs.items()
        })

    def visit_reference(self, const):
        return ConstValueReference(const.name, const.lineno)


const_value_or_ref = ConstValueMapper().get



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from .generate import Generator
from .link import TypeSpecLinker
from .link import ConstSpecLinker
from .link import ServiceSpecLinker
from .scope import Scope
from .exceptions import ThriftCompilerError


__all__ = ['Compiler']


class Compiler(object):
    """Compiles IDLs into Python modules."""

    LINKERS = [ConstSpecLinker, TypeSpecLinker, ServiceSpecLinker]

    __slots__ = ('protocol',)

    def __init__(self, protocol):
        """Initialize the compiler.

        :param thriftrw.protocol.Protocol protocol:
           The protocol ot use to serialize and deserialize values.
        """
        self.protocol = protocol

    def compile(self, name, program):
        """Compile the given parsed Thrift document into a Python module.

        The generated module contains,

        .. py:attribute:: services

            A collection of generated classes for all services defined in the
            thrift file.

        .. py:attribute:: types

            A collection of generated types for all types defined in the
            thrift file.

        .. py:attribute:: constants

            A mapping of constant name to value for all constants defined in
            the thrift file.

        .. py:function:: dumps(obj)

            Serializes the given object using the protocol the compiler was
            instantiated with.

        .. py:function:: loads(cls, payload)

            Deserializes an object of type ``cls`` from ``payload`` using the
            protocol the compiler was instantiated with.

        And one class each for every struct, union, exception, enum, and
        service defined in the IDL.

        Service classes have references to
        :py:class:`thriftrw.spec.ServiceFunction` objects for each method
        defined in the service.

        .. versionadded:: 0.2
           The ``constants`` attribute in generated modules.

        :param str name:
            Name of the Thrift document. This will be the name of the
            generated module.
        :param thriftrw.idl.Program program:
            AST of the parsted Thrift document.
        :returns:
            The generated module.
        """
        scope = Scope(name)

        for header in program.headers:
            header.apply(self)

        generator = Generator(scope)
        for definition in program.definitions:
            generator.process(definition)

        # TODO Linker can probably just be a callable.
        for linker in self.LINKERS:
            linker(scope).link()

        scope.add_function('loads', self.protocol.loads)
        scope.add_function('dumps', self.protocol.dumps)

        return scope.module

    def visit_include(self, include):
        raise ThriftCompilerError(
            'Include of "%s" found on line %d. '
            'thriftrw does not support including other Thrift files.'
            % (include.path, include.lineno)
        )

    def visit_namespace(self, namespace):
        pass  # nothing to do



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

import six
from ply import lex

from .exceptions import ThriftParserError


__all__ = ['Lexer']


THRIFT_KEYWORDS = (
    'namespace',
    'include',
    'void',
    'bool',
    'byte',
    'i16',
    'i32',
    'i64',
    'double',
    'string',
    'binary',
    'map',
    'list',
    'set',
    'oneway',
    'typedef',
    'struct',
    'union',
    'exception',
    'extends',
    'throws',
    'service',
    'enum',
    'const',
    'required',
    'optional',
)


class LexerSpec(object):
    """Lexer specification for Thrift IDL files.

    Adapted from thriftpy.parser.lexer."""

    literals = ':;,=*{}()<>[]'

    tokens = (
        'BOOLCONSTANT',
        'INTCONSTANT',
        'DUBCONSTANT',
        'LITERAL',
        'IDENTIFIER',
    ) + tuple(map(six.text_type.upper, THRIFT_KEYWORDS))

    t_ignore = ' \t\r'  # whitespace

    def t_error(self, t):
        raise ThriftParserError(
            'Illegal characher %r at line %d' % (t.value[0], t.lexer.lineno)
        )

    def t_newline(self, t):
        r'\n+'
        t.lexer.lineno += len(t.value)

    def t_ignore_SILLYCOMM(self, t):
        r'\/\*\**\*\/'
        t.lexer.lineno += t.value.count('\n')

    def t_ignore_MULTICOMM(self, t):
        r'\/\*[^*]\/*([^*/]|[^*]\/|\*[^/])*\**\*\/'
        t.lexer.lineno += t.value.count('\n')

    def t_ignore_DOCTEXT(self, t):
        r'\/\*\*([^*/]|[^*]\/|\*[^/])*\**\*\/'
        t.lexer.lineno += t.value.count('\n')

    def t_ignore_UNIXCOMMENT(self, t):
        r'\#[^\n]*'

    def t_ignore_COMMENT(self, t):
        r'\/\/[^\n]*'

    def t_BOOLCONSTANT(self, t):
        r'true|false'
        t.value = t.value == 'true'
        return t

    def t_DUBCONSTANT(self, t):
        r'-?\d+\.\d*(e-?\d+)?'
        t.value = float(t.value)
        return t

    def t_HEXCONSTANT(self, t):
        r'0x[0-9A-Fa-f]+'
        t.value = int(t.value, 16)
        t.type = 'INTCONSTANT'
        return t

    def t_INTCONSTANT(self, t):
        r'[+-]?[0-9]+'
        t.value = int(t.value)
        return t

    def t_LITERAL(self, t):
        r'(\"([^\\\n]|(\\.))*?\")|\'([^\\\n]|(\\.))*?\''
        s = t.value[1:-1]
        maps = {
            't': '\t',
            'r': '\r',
            'n': '\n',
            '\\': '\\',
            '\'': '\'',
            '"': '\"'
        }
        i = 0
        length = len(s)
        val = ''
        while i < length:
            if s[i] == '\\':
                i += 1
                if s[i] in maps:
                    val += maps[s[i]]
                else:
                    msg = 'Cannot escape character: %s' % s[i]
                    raise ThriftParserError(msg)
            else:
                val += s[i]
            i += 1

        t.value = val
        return t

    def t_IDENTIFIER(self, t):
        r'[a-zA-Z_](\.[a-zA-Z_0-9]|[a-zA-Z_0-9])*'

        if t.value in THRIFT_KEYWORDS:
            # Not an identifier after all.
            t.type = t.value.upper()

        return t


class Lexer(LexerSpec):
    """Lexer for Thrift IDL files."""

    def __init__(self, **kwargs):
        self._lexer = lex.lex(module=self, **kwargs)

    def input(self, data):
        """Reset the lexer and feed in new input.

        :param data:
            String of input data.
        """
        # input(..) doesn't reset the lineno. We have to do that manually.
        self._lexer.lineno = 1
        return self._lexer.input(data)

    def token(self):
        """Return the next token.

        Returns None when the end of the input is reached.
        """
        return self._lexer.token()



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Protocols
---------

.. autoclass:: thriftrw.protocol.Protocol
    :members:

.. autoclass:: thriftrw.protocol.BinaryProtocol
    :members:

Exceptions
----------

.. autoclass:: thriftrw.protocol.ThriftProtocolError
    :members:

.. autoclass:: thriftrw.protocol.EndOfInputError
    :members:
"""
from __future__ import absolute_import, unicode_literals, print_function

from .binary import BinaryProtocol
from .core import Protocol
from .exceptions import ThriftProtocolError, EndOfInputError


__all__ = [
    # Protocol
    'Protocol',
    'BinaryProtocol',

    # Exceptions
    'ThriftProtocolError',
    'EndOfInputError',
]



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

import collections

from thriftrw.wire import TType
from thriftrw.wire.value import SetValue

from . import check
from .base import TypeSpec

__all__ = ['SetTypeSpec']


class SetTypeSpec(TypeSpec):
    """
    :param TypeSpec vspec:
        TypeSpec of values stored in the set.
    """

    __slots__ = ('vspec', 'linked')

    ttype_code = TType.SET

    def __init__(self, vspec):
        self.vspec = vspec
        self.linked = False

    def link(self, scope):
        if not self.linked:
            self.linked = True
            self.vspec = self.vspec.link(scope)
        return self

    @property
    def name(self):
        return 'set<%s>' % self.vspec.name

    def to_wire(self, value):
        check.instanceof_class(self, collections.Set, value)
        return SetValue(
            value_ttype=self.vspec.ttype_code,
            values=[self.vspec.to_wire(v) for v in value],
        )

    def from_wire(self, wire_value):
        check.type_code_matches(self, wire_value)
        return set(
            self.vspec.from_wire(v) for v in wire_value.values
        )

    def __str__(self):
        return 'SetTypeSpec(vspec=%r)' % self.vspec

    __repr__ = __str__

    def __eq__(self, other):
        return (
            self.vspec == other.vspec and
            self.linked == other.linked
        )



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function


def fields_eq(fields):
    """Generates an ``__eq__`` method.

    :param list fields:
        List of fields of the object to be compared.
    """
    def __eq__(self, other):
        return all(
            getattr(self, name) == getattr(other, name) for name in fields
        )
    return __eq__


def fields_str(cls_name, field_list, include_none=True):
    """Generates a ``__str__`` method.

    :param cls_name:
        Name of the class for which the method is being generated.
    :param list field_list:
        List of field_list of the class to be included in the output.
    :param bool include_none:
        Whether None attributes should be included in the output.
    """
    def __str__(self):
        fields = {}

        for name in field_list:
            # TODO use to_primitive?
            value = getattr(self, name)
            if include_none or value is not None:
                fields[name] = value

        return "%s(%r)" % (cls_name, fields)
    return __str__



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

import six

from thriftrw.wire import TType
from thriftrw.wire.value import (
    BoolValue,
    ByteValue,
    DoubleValue,
    I16Value,
    I32Value,
    I64Value,
    BinaryValue,
)

from . import check
from .base import TypeSpec

__all__ = [
    'BoolTypeSpec',
    'ByteTypeSpec',
    'DoubleTypeSpec',
    'I16TypeSpec',
    'I32TypeSpec',
    'I64TypeSpec',
    'BinaryTypeSpec',
    'TextTypeSpec',
]


class PrimitiveTypeSpec(TypeSpec):
    """TypeSpec for primitive types."""

    __slots__ = ('name', 'code', 'value_cls', 'surface')

    def __init__(self, name, code, value_cls, surface):
        self.name = name
        self.code = code
        self.value_cls = value_cls
        self.surface = surface

    @property
    def ttype_code(self):
        return self.code

    def to_wire(self, value):
        check.instanceof_surface(self, value)
        # TODO check bounds for numeric values.
        return self.value_cls(value)

    def from_wire(self, wire_value):
        check.type_code_matches(self, wire_value)
        return wire_value.value

    def link(self, scope):
        return self

    def __str__(self):
        return 'PrimitiveType(%r, %s)' % (self.code, self.value_cls)

    __repr__ = __str__


class _TextTypeSpec(TypeSpec):
    """TypeSpec for the text type."""

    __slots__ = ()

    name = 'string'
    surface = six.text_type

    @property
    def ttype_code(self):
        return TType.BINARY

    def to_wire(self, value):
        if isinstance(value, six.text_type):
            value = value.encode('utf-8')
        elif not isinstance(value, six.binary_type):
            raise TypeError(
                'Cannot serialize %r into a "string".' % (value,)
            )
        return BinaryValue(value)

    def from_wire(self, wire_value):
        check.type_code_matches(self, wire_value)
        return wire_value.value.decode('utf-8')

    def link(self, scope):
        return self


#: TypeSpec for boolean values.
BoolTypeSpec = PrimitiveTypeSpec('bool', TType.BOOL, BoolValue, bool)

#: TypeSpec for single-byte integers.
ByteTypeSpec = PrimitiveTypeSpec('byte', TType.BYTE, ByteValue, int)

#: TypeSpec for floating point numbers with 64 bits of precision.
DoubleTypeSpec = PrimitiveTypeSpec('double', TType.DOUBLE, DoubleValue, float)

#: TypeSpec for 16-bit integers.
I16TypeSpec = PrimitiveTypeSpec('i16', TType.I16, I16Value, int)

#: TypeSpec for 32-bit integers.
I32TypeSpec = PrimitiveTypeSpec('i32', TType.I32, I32Value, int)

#: TypeSpec for 64-bit integers.
I64TypeSpec = PrimitiveTypeSpec('i64', TType.I64, I64Value, int)

#: TypeSpec for binary blobs.
BinaryTypeSpec = PrimitiveTypeSpec(
    'binary', TType.BINARY, BinaryValue, six.binary_type
)

#: TypeSpec for unicode data.
#:
#: Values will be decoded/encoded using UTF-8 encoding before/after being
#: serialized/deserialized.
#:
#: .. versionchanged:: 0.3.1
#:
#:     Allows passing binary values directly.
TextTypeSpec = _TextTypeSpec()



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Specs
-----

Specs represent compiled ASTs and act as the *specification* of behavior for
the types or values they represent. All specs have a ``name`` and may have a
``surface`` attribute which, if not None, is exposed in the generated module
at the top-level.

.. autoclass:: thriftrw.spec.TypeSpec
    :members:

Native types
~~~~~~~~~~~~

.. autodata:: thriftrw.spec.BoolTypeSpec
    :annotation:

.. autodata:: thriftrw.spec.ByteTypeSpec
    :annotation:

.. autodata:: thriftrw.spec.DoubleTypeSpec
    :annotation:

.. autodata:: thriftrw.spec.I16TypeSpec
    :annotation:

.. autodata:: thriftrw.spec.I32TypeSpec
    :annotation:

.. autodata:: thriftrw.spec.I64TypeSpec
    :annotation:

.. autodata:: thriftrw.spec.BinaryTypeSpec
    :annotation:

.. autodata:: thriftrw.spec.TextTypeSpec
    :annotation:

.. autoclass:: thriftrw.spec.ListTypeSpec
    :members:

.. autoclass:: thriftrw.spec.MapTypeSpec
    :members:

.. autoclass:: thriftrw.spec.SetTypeSpec
    :members:

Custom Types
~~~~~~~~~~~~

.. autoclass:: thriftrw.spec.EnumTypeSpec
    :members:

.. autoclass:: thriftrw.spec.StructTypeSpec
    :members:

.. autoclass:: thriftrw.spec.ExceptionTypeSpec
    :members:

.. autoclass:: thriftrw.spec.UnionTypeSpec
    :members:

.. autoclass:: thriftrw.spec.FieldSpec
    :members:

.. autoclass:: thriftrw.spec.TypedefTypeSpec
    :members:

Services
~~~~~~~~

.. autoclass:: thriftrw.spec.ServiceSpec
    :members:

.. autoclass:: thriftrw.spec.ServiceFunction
    :members:

.. autoclass:: thriftrw.spec.FunctionSpec
    :members:

.. autoclass:: thriftrw.spec.FunctionResultSpec
    :members:

Constants
~~~~~~~~~

.. autoclass:: thriftrw.spec.ConstSpec
    :members:
"""
from __future__ import absolute_import, unicode_literals, print_function

from .base import TypeSpec
from .const import ConstSpec
from .enum import EnumTypeSpec
from .list import ListTypeSpec
from .map import MapTypeSpec
from .set import SetTypeSpec
from .struct import StructTypeSpec, FieldSpec
from .exc import ExceptionTypeSpec
from .union import UnionTypeSpec
from .typedef import TypedefTypeSpec
from .service import (
    ServiceSpec,
    FunctionSpec,
    FunctionResultSpec,
    ServiceFunction,
)
from .primitive import (
    BoolTypeSpec,
    ByteTypeSpec,
    DoubleTypeSpec,
    I16TypeSpec,
    I32TypeSpec,
    I64TypeSpec,
    BinaryTypeSpec,
    TextTypeSpec,
)


__all__ = [
    # Types
    'TypeSpec',

    # Primitives
    'BoolTypeSpec',
    'ByteTypeSpec',
    'DoubleTypeSpec',
    'I16TypeSpec',
    'I32TypeSpec',
    'I64TypeSpec',
    'BinaryTypeSpec',
    'TextTypeSpec',

    # Containers
    'ListTypeSpec',
    'MapTypeSpec',
    'SetTypeSpec',

    # Custom types
    'EnumTypeSpec',

    'ExceptionTypeSpec',
    'StructTypeSpec',
    'UnionTypeSpec',
    'FieldSpec',

    'TypedefTypeSpec',

    # Services
    'ServiceSpec',
    'FunctionSpec',
    'FunctionResultSpec',
    'ServiceFunction',

    # Constants
    'ConstSpec',
]



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

import abc

__all__ = ['TypeSpec']


class TypeSpec(object):
    """Base class for classes representing TypeSpecs.

    A TypeSpec knows how to convert values of the corresponding type to and
    from :py:class:`thriftrw.wire.Value` objects.
    """
    __metaclass__ = abc.ABCMeta

    @property
    def name(self):
        """Name of the type referenced by this type spec."""
        raise NotImplementedError

    @property
    def ttype_code(self):
        """Numeric TType used for the type spec.

        The value must be from :py:data:`thriftrw.wire.TType`.
        """
        raise NotImplementedError

    @property
    def surface(self):
        """The surface of a type spec.

        The surface of a type spec, if non-None is the value associated with
        its name at the top-level in the generated module.
        """
        raise NotImplementedError

    @abc.abstractmethod
    def to_wire(self, value):
        """Converts the given value into a :py:class:`thriftrw.wire.Value`
        object.

        :returns thriftrw.wire.Value:
            Wire representation of the value.
        :raises TypeError:
            If the value did not match the type expected by this TypeSpec.
        :raises ValueError:
            If the value matched the type but did not hold the correct set of
            acceptable values.
        """

    @abc.abstractmethod
    def from_wire(self, wire_value):
        """Converts the given :py:class:`thriftrw.wire.Value` back into the
        original type.

        :param thriftr.wire.Value wire_value:
            Value to convert.
        :raises ValueError:
            If the type of the wire value does not have the correct Thrift
            type for this type spec.
        """

    @abc.abstractmethod
    def link(self, scope):
        pass



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
Types
-----

.. autodata:: thriftrw.wire.TType
    :annotation:

Value
-----

.. autoclass:: thriftrw.wire.Value
    :members:

.. autoclass:: thriftrw.wire.ValueVisitor
    :members:

Value Types
~~~~~~~~~~~

.. autoclass:: thriftrw.wire.BoolValue

.. autoclass:: thriftrw.wire.ByteValue

.. autoclass:: thriftrw.wire.DoubleValue

.. autoclass:: thriftrw.wire.I16Value

.. autoclass:: thriftrw.wire.I32Value

.. autoclass:: thriftrw.wire.I64Value

.. autoclass:: thriftrw.wire.BinaryValue

.. autoclass:: thriftrw.wire.FieldValue

.. autoclass:: thriftrw.wire.StructValue

.. autoclass:: thriftrw.wire.MapValue

.. autoclass:: thriftrw.wire.SetValue

.. autoclass:: thriftrw.wire.ListValue
"""
from __future__ import absolute_import, unicode_literals, print_function

from .ttype import TType

from .value import (
    Value,
    BoolValue,
    ByteValue,
    DoubleValue,
    I16Value,
    I32Value,
    I64Value,
    BinaryValue,
    FieldValue,
    StructValue,
    MapValue,
    SetValue,
    ListValue,
    ValueVisitor,
)

__all__ = [
    'TType',

    # Value
    'Value',
    'BoolValue',
    'ByteValue',
    'DoubleValue',
    'I16Value',
    'I32Value',
    'I64Value',
    'BinaryValue',
    'FieldValue',
    'StructValue',
    'MapValue',
    'SetValue',
    'ListValue',
    'ValueVisitor',
]



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

import types

from .exceptions import ThriftCompilerError

__all__ = ['Scope']


class Scope(object):
    """Maintains the compilation state across steps.

    The scope is not exported to the user directly. It's only used to maintain
    state of known types and values during the compilation process and holds a
    reference to the final generated module.
    """

    __slots__ = ('const_specs', 'type_specs', 'module', 'service_specs')

    def __init__(self, name):
        """Initialize the scope.

        :param name:
            Name of the generated module.
        """
        self.type_specs = {}
        self.const_specs = {}
        self.service_specs = {}

        self.module = types.ModuleType(str(name))

    def __str__(self):
        return "Scope(%r)" % {
            'const_values': self.const_values,
            'type_specs': self.type_specs,
            'service_specs': self.service_specs,
            'module': self.module,
        }

    __repr__ = __str__

    def resolve_const_spec(self, name, lineno):
        """Finds and links the ConstSpec with the given name."""
        if name not in self.const_specs:
            raise ThriftCompilerError(
                'Unknown constant "%s" referenced at line %d' % (name, lineno)
            )
        return self.const_specs[name].link(self)

    def resolve_type_spec(self, name, lineno):
        """Finds and links the TypeSpec with the given name."""
        if name not in self.type_specs:
            raise ThriftCompilerError(
                'Unknown type "%s" referenced at line %d' % (name, lineno)
            )
        return self.type_specs[name].link(self)

    def add_service_spec(self, service_spec):
        """Registers the given ``ServiceSpec`` into the scope.

        Raises ``ThriftCompilerError`` if the name has already been used.
        """
        assert service_spec is not None

        if service_spec.name in self.service_specs:
            raise ThriftCompilerError(
                'Cannot define service "%s". That name is already taken.'
                % service_spec.name
            )
        self.service_specs[service_spec.name] = service_spec

    def add_const_spec(self, const_spec):
        """Adds a ConstSpec to the compliation scope.

        If the ConstSpec's ``save`` attribute is True, the constant will be
        added to the module at the top-level.
        """
        if const_spec.name in self.const_specs:
            raise ThriftCompilerError(
                'Cannot define constant "%s". That name is already taken.'
                % const_spec.name
            )
        self.const_specs[const_spec.name] = const_spec

    def add_surface(self, name, surface):
        assert surface is not None

        if hasattr(self.module, name):
            raise ThriftCompilerError(
                'Cannot define "%s". The name has already been used.' % name
            )

        setattr(self.module, name, surface)

    def add_function(self, name, func):
        """Adds a top-level function with the given name to the module.

        :param name:
            Name of the function.
        :param func:
            Function to add to the module.
        """
        assert func is not None

        if hasattr(self.module, name):
            raise ThriftCompilerError(
                'Cannot define "%s". The name has already been used.' % name
            )

        setattr(self.module, name, func)

    def add_type_spec(self, name, spec, lineno):
        """Adds the given type to the scope.

        :param str name:
            Name of the new type
        :param spec:
            ``TypeSpec`` object containing information on the type, or a
            ``TypeReference`` if this is meant to be resolved during the
            ``link`` stage.
        :param lineno:
            Line number on which this type is defined.
        """
        assert type is not None

        if name in self.type_specs:
            raise ThriftCompilerError(
                'Cannot define type "%s" at line %d. '
                'Another type with that name already exists.'
                % (name, lineno)
            )

        self.type_specs[name] = spec



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function


__all__ = ['type_code_matches', 'instanceof_surface']


def type_code_matches(type_spec, wire_value):
    """Verifies that the ttype_code for the TypeSpec and the Value matches.

    Raises ``ValueError`` if not.

    :param thriftrw.spec.base.TypeSpec type_spec:
        TypeSpec against which the value is being checked.
    :param thriftrw.idl.Value wire_value:
        Wire value being checked.
    """
    if type_spec.ttype_code != wire_value.ttype_code:
        raise ValueError(
            'Cannot read a "%s" (type %d) from %r (type %d)' %
            (type_spec.name, type_spec.ttype_code, wire_value,
             wire_value.ttype_code)
        )


def instanceof_surface(type_spec, value):
    """Verifies that the value is an instance of the type's surface.

    Raises ``TypeError`` if not.

    :param thriftrw.spec.base.TypeSpec type_spec:
        TypeSpec against which the value is being checked.
    :param value:
        Value being validated.
    """
    instanceof_class(type_spec, type_spec.surface, value)


def instanceof_class(type_spec, cls, value):
    if not isinstance(value, cls):
        raise TypeError(
            'Cannot serialize %r into a "%s".' % (value, type_spec.name)
        )



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from collections import namedtuple

__all__ = ['TType']

_TType = namedtuple(
    '_TType',
    'BOOL BYTE DOUBLE I16 I32 I64 BINARY STRUCT MAP SET LIST'
)

#: Different TType codes supported by Thrift.
#:
#: .. autoattribute:: TType.BOOL
#:
#: .. autoattribute:: TType.BYTE
#:
#: .. autoattribute:: TType.DOUBLE
#:
#: .. autoattribute:: TType.I16
#:
#: .. autoattribute:: TType.I32
#:
#: .. autoattribute:: TType.I64
#:
#: .. autoattribute:: TType.BINARY
#:
#: .. autoattribute:: TType.STRUCT
#:
#: .. autoattribute:: TType.MAP
#:
#: .. autoattribute:: TType.SET
#:
#: .. autoattribute:: TType.LIST
#:
TType = _TType(
    BOOL=2,
    BYTE=3,
    DOUBLE=4,
    I16=6,
    I32=8,
    I64=10,
    BINARY=11,
    STRUCT=12,
    MAP=13,
    SET=14,
    LIST=15,
)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from thriftrw.wire import TType
from thriftrw.wire.value import StructValue
from thriftrw.compile.exceptions import ThriftCompilerError

from . import check
from . import common
from .base import TypeSpec
from .struct import FieldSpec

__all__ = ['UnionTypeSpec', 'FieldSpec']


class UnionTypeSpec(TypeSpec):
    """Spec for Thrift unions.

    The surface for union types is a class with the following:

    .. py:attribute:: type_spec

        :py:class:`UnionTypeSpec` for the type.

    .. py:method:: __init__(self, *args, **kwargs)

        Accepts all fields of the unions as keyword arguments but only one of
        them is allowed to be non-None. Positional arguments are not accepted.

    And obvious definitions of ``__str__`` and ``__eq__``.

    Given the definition,::

        union Body {
            1: string plainText
            2: binary richText
        }

    A class roughly equivalent to the following is generated,

    .. code-block:: python

        class Body(object):

            __slots__ = ('plainText', 'richText')

            def __init__(self, plainText=None, richText=None):
                # Only one of plainText and richText may be non-None.
                # ...
    """

    __slots__ = ('name', 'fields', 'linked', 'surface', 'allow_empty')

    ttype_code = TType.STRUCT

    def __init__(self, name, fields, allow_empty=None):
        self.name = name
        self.fields = fields
        self.linked = False
        self.surface = None
        self.allow_empty = allow_empty

    def link(self, scope):
        if not self.linked:
            self.linked = True
            self.fields = [field.link(scope) for field in self.fields]
            self.surface = union_cls(self, scope)
        return self

    @classmethod
    def compile(cls, union):
        fields = []
        ids = set()
        names = set()
        for field in union.fields:
            if field.name in names:
                raise ThriftCompilerError(
                    'Field "%s" of union "%s" on line %d has duplicates.' %
                    (field.name, union.name, field.lineno)
                )
            names.add(field.name)

            if field.id in ids:
                raise ThriftCompilerError(
                    'Field ID "%d" of union "%s" on line %d '
                    'has already been used.' % (
                        field.id, union.name, field.lineno
                    )
                )
            ids.add(field.id)

            if field.requiredness is not None:
                raise ThriftCompilerError(
                    'Field "%s" of union "%s" on line %d is "%s". '
                    'Unions cannot specify requiredness. '
                    % (
                        field.name,
                        union.name,
                        field.lineno,
                        'required' if field.requiredness else 'optional',
                    )
                )

            if field.default is not None:
                raise ThriftCompilerError(
                    'Field "%s" of union "%s" on line %d has a default value.'
                    ' Fields of unions cannot have default values. '
                    % (field.name, union.name, field.lineno)
                )

            fields.append(FieldSpec.compile(
                field=field,
                struct_name=union.name,
                require_requiredness=False,
            ))
        return cls(union.name, fields)

    def to_wire(self, union):
        check.instanceof_surface(self, union)
        fields = []

        for field in self.fields:
            value = getattr(union, field.name)
            if value is None:
                continue
            fields.append(field.to_wire(value))

        return StructValue(fields)

    def from_wire(self, wire_value):
        check.type_code_matches(self, wire_value)
        kwargs = {}
        for field in self.fields:
            field_value = wire_value.get(field.id, field.spec.ttype_code)
            if field_value is None:
                continue
            kwargs[field.name] = field.from_wire(field_value)

        # TODO For the case where cls fails to instantiate because a required
        # positional argument is missing, we know that the request was
        # invalid.
        return self.surface(**kwargs)

    def __str__(self):
        return 'UnionTypeSpec(name=%r, fields=%r)' % (self.name, self.fields)

    __repr__ = __str__

    def __eq__(self, other):
        return (
            self.name == other.name and
            self.fields == other.fields and
            self.linked == other.linked
        )


def union_init(cls_name, fields, allow_empty):
    """Generates the ``__init__`` method for unions.

    :param cls_name:
        Name of the class.
    :param fields:
        Collection of fields on the union.
    :param allow_empty:
        Whether a union with no values assigned is allowed.
    """

    def __init__(self, *args, **kwargs):
        if args:
            raise TypeError(
                '%s() does not accept any positional arguments (%d given)'
                % (cls_name, len(args))
            )

        assigned = None

        for field in fields:
            value = None

            if field in kwargs:
                value = kwargs.pop(field)

            setattr(self, field, value)

            if value is not None:
                if assigned is not None:
                    raise TypeError(
                        '%s() received multiple values, "%s" and "%s". '
                        'Unions can have at most one field populated. '
                        % (cls_name, assigned, field)
                    )
                else:
                    assigned = field

        if kwargs:
            # Too many kwargs given.
            for name in kwargs:
                raise TypeError(
                    '%s() got an unexpected keyword argument "%s"'
                    % (cls_name, name)
                )

        if assigned is None and fields and not allow_empty:
            raise TypeError(
                '%s() did not receive any values. '
                'Exactly one non-None value is required.'
                % cls_name
            )

    return __init__


def union_docstring(cls_name, fields):
    """Generates a docstring for generated unions."""

    fields_section = (
        '%s is a union of the following fields. '
        'Exactly one of them must be provided.\n\n' % cls_name
    ) + '\n'.join('-   %s' % name for name in fields)

    params = ', '.join('%s=None' % name for name in fields)
    header = '%s(%s)' % (cls_name, params)

    return header + '\n\n' + fields_section


def union_cls(union_spec, scope):
    """Generates a class for a union.

    The generated class has the following attributes:

    ``type_spec``
        Attribute pointing to the ``TypeSpec`` for the type.

    The ``__init__`` method of the class accepts all fields of the union as
    keyword arguments, and allows only one of them to be non-None.

    A ``__str__`` method that returns which field is populated and with what
    value is also added to the class.

    :param UnionTypeSpec union_spec:
        Spec of the union.
    :param thriftrw.compile.Scope scope:
        Compilation scope
    """
    field_names = set(field.name for field in union_spec.fields)

    union_dct = {}
    union_dct['type_spec'] = union_spec
    union_dct['__slots__'] = tuple(field_names)
    union_dct['__init__'] = union_init(
        union_spec.name, field_names, union_spec.allow_empty
    )
    union_dct['__str__'] = common.fields_str(
        union_spec.name, field_names, False
    )
    union_dct['__repr__'] = union_dct['__str__']
    union_dct['__eq__'] = common.fields_eq(field_names)
    union_dct['__doc__'] = union_docstring(
        union_spec.name, field_names
    )

    return type(str(union_spec.name), (object,), union_dct)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

import six
import struct
from six.moves import range

from thriftrw.wire import value as V
from thriftrw.wire import TType

from .core import Protocol
from .exceptions import EndOfInputError
from .exceptions import ThriftProtocolError


STRUCT_END = 0


class BinaryProtocolReader(object):
    """Parser for the binary protocol."""

    __slots__ = ('reader',)

    def __init__(self, reader):
        """Initialize the reader.

        :param reader:
            File-like object with a ``read(num)`` method which returns
            *exactly* the requested number of bytes.
        """
        self.reader = reader

    def read(self, typ):
        return self._reader(typ)(self)

    def _read(self, num_bytes):
        chunk = self.reader.read(num_bytes)
        if len(chunk) != num_bytes:
            raise EndOfInputError(
                'Expected %d bytes but got %d bytes.' %
                (num_bytes, len(chunk))
            )
        return chunk

    def _unpack(self, num_bytes, spec):
        """Unpack a single value using the given format and return it.

        :param num_bytes:
            Number of bytes to read.
        :param spec:
            Spec for ``struct.unpack``
        """
        (result,) = struct.unpack(spec, self._read(num_bytes))
        return result

    def _byte(self):
        return self._unpack(1, '!b')

    def _i16(self):
        return self._unpack(2, '!h')

    def _i32(self):
        return self._unpack(4, '!i')

    def read_bool(self):
        """Reads a boolean."""
        return V.BoolValue(self._byte() == 1)

    def read_byte(self):
        """Reads a byte."""
        return V.ByteValue(self._byte())

    def read_double(self):
        """Reads a double."""
        return V.DoubleValue(self._unpack(8, '!d'))

    def read_i16(self):
        """Reads a 16-bit integer."""
        return V.I16Value(self._i16())

    def read_i32(self):
        """Reads a 32-bit integer."""
        return V.I32Value(self._i32())

    def read_i64(self):
        """Reads a 64-bit integer."""
        return V.I64Value(self._unpack(8, '!q'))

    def read_binary(self):
        """Reads a binary blob."""
        length = self._i32()
        return V.BinaryValue(self._read(length))

    def read_struct(self):
        """Reads an arbitrary Thrift struct."""
        fields = []

        field_type = self._byte()
        while field_type != STRUCT_END:
            field_id = self._i16()
            field_value = self.read(field_type)
            fields.append(
                V.FieldValue(
                    id=field_id,
                    ttype=field_type,
                    value=field_value,
                )
            )

            field_type = self._byte()
        return V.StructValue(fields)

    def read_map(self):
        """Reads a map."""
        key_ttype = self._byte()
        value_ttype = self._byte()
        length = self._i32()

        key_reader = self._reader(key_ttype)
        value_reader = self._reader(value_ttype)

        pairs = []
        for i in range(length):
            k = key_reader(self)
            v = value_reader(self)
            pairs.append((k, v))

        return V.MapValue(
            key_ttype=key_ttype,
            value_ttype=value_ttype,
            pairs=pairs,
        )

    def read_set(self):
        """Reads a set."""
        value_ttype = self._byte()
        length = self._i32()

        value_reader = self._reader(value_ttype)
        values = []

        for i in range(length):
            values.append(value_reader(self))

        return V.SetValue(
            value_ttype=value_ttype,
            values=values
        )

    def read_list(self):
        """Reads a list."""
        value_ttype = self._byte()
        length = self._i32()

        value_reader = self._reader(value_ttype)
        values = []

        for i in range(length):
            values.append(value_reader(self))

        return V.ListValue(
            value_ttype=value_ttype,
            values=values
        )

    def _reader(self, typ):
        reader = self._readers.get(typ)
        if reader is None:
            raise ThriftProtocolError('Unknown TType "%r"' % typ)
        return reader

    # Mapping of TType to function that can read it.
    _readers = {
        TType.BOOL: read_bool,
        TType.BYTE: read_byte,
        TType.DOUBLE: read_double,
        TType.I16: read_i16,
        TType.I32: read_i32,
        TType.I64: read_i64,
        TType.BINARY: read_binary,
        TType.STRUCT: read_struct,
        TType.MAP: read_map,
        TType.SET: read_set,
        TType.LIST: read_list,
    }


class BinaryProtocolWriter(V.ValueVisitor):
    """Serializes values using the Thrift Binary protocol."""

    __slots__ = ('writer',)

    def __init__(self, writer):
        """Initialize the writer.

        :param writer:
            File-like object with ``write`` method.
        """
        self.writer = writer

    def write(self, value):
        """Writes the given value.

        :param value:
            A ``Value`` object.
        """
        value.apply(self)

    def _pack(self, spec, value):
        """Pack the given value using ``struct.pack`` and write it.

        :param spec:
            Spec to pass to ``struct.pack``.
        :param value:
            Value to pack.
        """
        self.writer.write(struct.pack(spec, value))

    def visit_bool(self, value):  # bool:1
        self.visit_byte(1 if value else 0)

    def visit_byte(self, value):  # byte:1
        self._pack('!b', value)

    def visit_double(self, value):  # double:8
        self._pack('!d', value)

    def visit_i16(self, value):  # i16:2
        self._pack('!h', value)

    def visit_i32(self, value):  # i32:4
        self._pack('!i', value)

    def visit_i64(self, value):  # i64:8
        self._pack('!q', value)

    def visit_binary(self, value):  # len:4 str:len
        self.visit_i32(len(value))
        self.writer.write(value)

    def visit_struct(self, fields):
        # ( type:1 id:2 value:* ){fields} '0'
        for field in fields:
            self.visit_byte(field.ttype)
            self.visit_i16(field.id)
            self.write(field.value)
        self.visit_byte(STRUCT_END)

    def visit_map(self, key_ttype, value_ttype, pairs):
        # key_type:1 value_type:1 count:4 (key:* value:*){count}
        self.visit_byte(key_ttype)
        self.visit_byte(value_ttype)
        self.visit_i32(len(pairs))

        for k, v in pairs:
            self.write(k)
            self.write(v)

    def visit_set(self, value_ttype, values):
        # value_type:1 count:4 (item:*){count}
        self.visit_byte(value_ttype)
        self.visit_i32(len(values))

        for v in values:
            self.write(v)

    def visit_list(self, value_ttype, values):
        # value_type:1 count:4 (item:*){count}
        self.visit_byte(value_ttype)
        self.visit_i32(len(values))

        for v in values:
            self.write(v)


class BinaryProtocol(Protocol):
    """Implements the Thrift binary protocol."""

    __slots__ = ()

    def serialize_value(self, value):
        buff = six.BytesIO()
        writer = BinaryProtocolWriter(buff)
        writer.write(value)
        return buff.getvalue()

    def deserialize_value(self, typ, s):
        buff = six.BytesIO(s)
        reader = BinaryProtocolReader(buff)
        return reader.read(typ)

__all__ = ['BinaryProtocol']



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from thriftrw.wire import TType
from thriftrw.wire.value import I32Value
from thriftrw.compile.exceptions import ThriftCompilerError

from . import check
from .base import TypeSpec

__all__ = ['EnumTypeSpec']


class EnumTypeSpec(TypeSpec):
    """TypeSpec for enum types.

    The surface for enum types is a class with the following:

    .. py:attribute:: type_spec

        :py:class:`EnumTypeSpec` for the type.

    .. py:attribute:: items

        A tuple of the names of all items in this enum.

    .. py:attribute:: values

        A tuple of the values of all items in this enum in the same order as
        ``items``.

    .. py:classmethod:: name_of(value)

        Finds the name of an enum item by its value.

        :param int value:
            A value for an item defined in this enum.
        :returns:
            Name of the item with that value or None if no such item exists.

    And one attribute for each enum item that has the same name as the
    attribute and points to the value for that item.

    Given the definition,::

        enum Role {
            User, Admin
        }

    The generated class is roughly equivalent to,

    .. code-block:: python

        class Role(object):
            User = 0
            Admin = 1

            items = ('User', 'Admin')
            values = (0, 1)

            type_spec = # ...

            def name_of(self, value):
                # ...
    """

    __slots__ = ('name', 'items', 'values_to_names', 'linked', 'surface')

    ttype_code = TType.I32

    def __init__(self, name, items):
        assert name
        assert items is not None

        #: Name of the enum class.
        self.name = name

        #: Mapping of enum item names to item values.
        self.items = items

        values_to_names = {}
        for name, value in items.items():
            if value in values_to_names:
                dupe = values_to_names[value]
                raise ThriftCompilerError(
                    'Items "%s" and "%s" of enum "%s" have value "%d". '
                    'Enums items cannot share values.' % (
                        name, dupe, self.name, value
                    )
                )
            values_to_names[value] = name

        #: Mapping of enum item values to their names.
        self.values_to_names = values_to_names
        self.linked = False

        #: The surface for this spec.
        self.surface = None

    def link(self, scope):
        if not self.linked:
            self.linked = True
            self.surface = enum_cls(self)
        return self

    def to_wire(self, value):
        if value not in self.values_to_names:
            raise ValueError(
                '%r is not a valid value for enum "%s"' % (value, self.name)
            )
        return I32Value(value)

    def from_wire(self, wire_value):
        check.type_code_matches(self, wire_value)
        return wire_value.value

    @classmethod
    def compile(cls, enum):
        items = {}

        # TODO check for int32 overflow.

        prev = -1
        for item in enum.items:
            value = item.value
            if value is None:
                value = prev + 1
            prev = value

            if item.name in items:
                raise ThriftCompilerError(
                    'Enum entry "%s.%s" on line %d has duplicates.' %
                    (enum.name, item.name, item.lineno)
                )

            items[item.name] = value

        return cls(enum.name, items)

    def __eq__(self, other):
        return (
            self.name == other.name and
            self.items == other.items and
            self.linked == other.linked
        )

    def __str__(self):
        return 'EnumTypeSpec(name=%r, items=%r)' % (self.name, self.items)

    __repr__ = __str__


def enum_cls(enum_spec):
    """Generates a class for the given EnumTypeSpec.

    :param EnumTypeSpec enum_spec:
        EnumTypeSpec for which the class is being generated.
    """
    enum_dct = {}

    for name, item in enum_spec.items.items():
        enum_dct[name] = item

    keys, values = zip(*enum_spec.items.items())

    enum_dct['_VALUES_TO_NAMES'] = enum_spec.values_to_names
    enum_dct['items'] = tuple(keys)
    enum_dct['values'] = tuple(values)
    enum_dct['name_of'] = name_of
    enum_dct['type_spec'] = enum_spec
    enum_dct['__slots__'] = ()
    enum_dct['__doc__'] = enum_docstring(
        enum_spec.name, enum_spec.items.items()
    )

    return type(str(enum_spec.name), (object,), enum_dct)


@classmethod
def name_of(cls, value):
    """Returns the name of the enum item with the given value.

    :param int value:
        Enum item value
    :returns:
        Name of the corresponding enum item or None if no such item exists.
    """
    return cls._VALUES_TO_NAMES.get(value)


def enum_docstring(name, items):
    item_section = '\n'.join(
        '-    %s = %s' % pair for pair in items
    )
    return '%s defines the following items:\n\n' % name + item_section



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from .spec_mapper import type_spec_or_ref


class TypedefTypeSpec(object):
    """Typedefs are aliases for other types.

    Typedefs resolve themselves to the target type at link time and eliminate
    themselves from the tree.
    """

    __slots__ = ('name', 'target_spec')

    def __init__(self, name, target_spec):
        self.name = name
        self.target_spec = target_spec

    @classmethod
    def compile(cls, typedef):
        target_spec = type_spec_or_ref(typedef.target_type)
        return cls(typedef.name, target_spec)

    def link(self, scope):
        return self.target_spec.link(scope)

    def __str__(self):
        return 'TypedefTypeSpec(%r, %r)' % (self.name, self.target_spec)

    __repr__ = __str__

    def __eq__(self, other):
        return (
            self.name == other.name and
            self.target_spec == other.target_spec
        )



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

"""
A parser for Thrift IDL files.

Parser
------

Adapted from ``thriftpy.parser``.

.. autoclass:: Parser
    :members:

.. autoclass:: ThriftParserError

AST
---

.. autoclass:: Program

Headers
~~~~~~~

.. autoclass:: Include

.. autoclass:: Namespace

Definitions
~~~~~~~~~~~

.. autoclass:: Const

.. autoclass:: Typedef

.. autoclass:: Enum

.. autoclass:: EnumItem

.. autoclass:: Struct

.. autoclass:: Union

.. autoclass:: Exc

.. autoclass:: Service

.. autoclass:: Function

.. autoclass:: Field

Types
~~~~~

.. autoclass:: PrimitiveType

.. autoclass:: MapType

.. autoclass:: SetType

.. autoclass:: ListType

.. autoclass:: DefinedType

Constants
~~~~~~~~~

.. autoclass:: ConstPrimitiveValue

.. autoclass:: ConstReference

Annotations
~~~~~~~~~~~

.. autoclass:: Annotation

Exceptions
----------

.. autoclass:: ThriftParserError
    :members:

"""
from __future__ import absolute_import, unicode_literals, print_function

from .exceptions import ThriftParserError
from .parser import Parser
from .ast import (
    Program,
    Include,
    Namespace,
    Const,
    Typedef,
    Enum,
    EnumItem,
    Struct,
    Union,
    Exc,
    Service,
    Function,
    Field,
    PrimitiveType,
    MapType,
    SetType,
    ListType,
    DefinedType,
    ConstValue,
    ConstPrimitiveValue,
    ConstReference,
    Annotation,
)


__all__ = [
    # Parser
    'Parser',

    # AST
    'Program',
    'Include',
    'Namespace',
    'Const',
    'Typedef',
    'Enum',
    'EnumItem',
    'Struct',
    'Union',
    'Exc',
    'Service',
    'Function',
    'Field',
    'PrimitiveType',
    'MapType',
    'SetType',
    'ListType',
    'DefinedType',
    'ConstValue',
    'ConstPrimitiveValue',
    'ConstReference',
    'Annotation',

    # Exceptions
    'ThriftParserError',
]



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from collections import deque

from ply import yacc

from . import ast
from .lexer import Lexer
from .exceptions import ThriftParserError


__all__ = ['Parser']


class ParserSpec(object):
    """Parser specification for Thrift IDL files.

    Adapted from ``thriftpy.parser.parser``."""

    tokens = Lexer.tokens

    def p_error(self, p):
        if p is None:
            raise ThriftParserError('Grammer error at EOF')
        raise ThriftParserError(
            'Grammar error %r at line %d' % (p.value, p.lineno)
        )

    def p_start(self, p):
        '''start : header definition'''
        p[0] = ast.Program(headers=p[1], definitions=p[2])

    def p_header(self, p):
        '''header : header_unit_ header
                  |'''
        self._parse_seq(p)

    def p_header_unit_(self, p):
        '''header_unit_ : header_unit ';'
                        | header_unit'''
        p[0] = p[1]

    def p_header_unit(self, p):
        '''header_unit : include
                       | namespace'''
        p[0] = p[1]

    def p_include(self, p):
        '''include : INCLUDE LITERAL'''
        p[0] = ast.Include(path=p[2], lineno=p.lineno(1))

    def p_namespace(self, p):
        '''namespace : NAMESPACE namespace_scope IDENTIFIER'''
        p[0] = ast.Namespace(scope=p[2], name=p[3], lineno=p.lineno(1))

    def p_namespace_scope(self, p):
        '''namespace_scope : '*'
                           | IDENTIFIER'''
        p[0] = p[1]

    def p_sep(self, p):
        '''sep : ','
               | ';'
        '''

    def p_definition(self, p):
        '''definition : definition_unit_ definition
                      |'''
        self._parse_seq(p)

    def p_definition_unit_(self, p):
        '''definition_unit_ : definition_unit ';'
                            | definition_unit'''
        p[0] = p[1]

    def p_definition_unit(self, p):
        '''definition_unit : const
                           | ttype
        '''
        p[0] = p[1]

    def p_const(self, p):
        '''const : CONST field_type IDENTIFIER '=' const_value
                 | CONST field_type IDENTIFIER '=' const_value sep'''
        p[0] = ast.Const(
            name=p[3],
            value_type=p[2],
            value=p[5],
            lineno=p.lineno(3),
        )

    def p_const_value(self, p):
        '''const_value : const_value_native
                       | const_ref'''
        p[0] = p[1]

    def p_const_value_native(self, p):
        '''const_value_native : const_value_primitive
                              | const_list
                              | const_map'''
        p[0] = p[1]

    def p_const_value_primitive(self, p):
        '''const_value_primitive : INTCONSTANT
                                 | DUBCONSTANT
                                 | LITERAL
                                 | BOOLCONSTANT'''
        p[0] = ast.ConstPrimitiveValue(p[1], lineno=p.lineno(1))

    def p_const_list(self, p):
        '''const_list : '[' const_list_seq ']' '''
        p[0] = ast.ConstList(list(p[2]), p.lineno(2))

    def p_const_list_seq(self, p):
        '''const_list_seq : const_value sep const_list_seq
                          | const_value const_list_seq
                          |'''
        self._parse_seq(p)

    def p_const_map(self, p):
        '''const_map : '{' const_map_seq '}' '''
        p[0] = ast.ConstMap(dict(p[2]), p.lineno(2))

    def p_const_map_seq(self, p):
        '''const_map_seq : const_map_item sep const_map_seq
                         | const_map_item const_map_seq
                         |'''
        self._parse_seq(p)

    def p_const_map_item(self, p):
        '''const_map_item : const_value ':' const_value '''
        p[0] = (p[1], p[3])

    def p_const_ref(self, p):
        '''const_ref : IDENTIFIER'''
        p[0] = ast.ConstReference(p[1], lineno=p.lineno(1))

    def p_ttype(self, p):
        '''ttype : typedef
                 | enum
                 | struct
                 | union
                 | exception
                 | service'''
        p[0] = p[1]

    def p_typedef(self, p):
        '''typedef : TYPEDEF field_type IDENTIFIER annotations'''
        p[0] = ast.Typedef(
            name=p[3], target_type=p[2], annotations=p[4], lineno=p.lineno(3)
        )

    def p_enum(self, p):  # noqa
        '''enum : ENUM IDENTIFIER '{' enum_seq '}' annotations'''
        p[0] = ast.Enum(
            name=p[2], items=p[4], annotations=p[6], lineno=p.lineno(2)
        )

    def p_enum_seq(self, p):
        '''enum_seq : enum_item sep enum_seq
                    | enum_item enum_seq
                    |'''
        self._parse_seq(p)

    def p_enum_item(self, p):
        '''enum_item : IDENTIFIER '=' INTCONSTANT annotations
                     | IDENTIFIER annotations'''
        if len(p) == 5:
            p[0] = ast.EnumItem(
                name=p[1], value=p[3], annotations=p[4], lineno=p.lineno(1)
            )
        else:
            p[0] = ast.EnumItem(
                name=p[1], value=None, annotations=p[2], lineno=p.lineno(1)
            )

    def p_struct(self, p):
        '''struct : STRUCT IDENTIFIER '{' field_seq '}' annotations'''
        p[0] = ast.Struct(
            name=p[2], fields=p[4], annotations=p[6], lineno=p.lineno(2)
        )

    def p_union(self, p):
        '''union : UNION IDENTIFIER '{' field_seq '}' annotations'''
        p[0] = ast.Union(
            name=p[2], fields=p[4], annotations=p[6], lineno=p.lineno(2)
        )

    def p_exception(self, p):
        '''exception : EXCEPTION IDENTIFIER '{' field_seq '}' annotations'''
        p[0] = ast.Exc(
            name=p[2], fields=p[4], annotations=p[6], lineno=p.lineno(2)
        )

    def p_service(self, p):
        '''service : SERVICE IDENTIFIER '{' function_seq '}' annotations
                   | SERVICE IDENTIFIER EXTENDS IDENTIFIER \
                     '{' function_seq '}' annotations
        '''

        if len(p) == 7:
            p[0] = ast.Service(
                name=p[2],
                functions=p[4],
                parent=None,
                annotations=p[6],
                lineno=p.lineno(2),
            )
        else:
            p[0] = ast.Service(
                name=p[2],
                functions=p[6],
                parent=p[4],
                annotations=p[8],
                lineno=p.lineno(2),
            )

    def p_oneway(self, p):
        '''oneway : ONEWAY
                  |'''
        p[0] = len(p) > 1

    def p_function(self, p):
        '''function : oneway function_type IDENTIFIER '(' field_seq ')' \
                      throws annotations '''
        p[0] = ast.Function(
            name=p[3],
            parameters=p[5],
            return_type=p[2],
            exceptions=p[7],
            oneway=p[1],
            annotations=p[8],
            lineno=p.lineno(3),
        )

    def p_function_seq(self, p):
        '''function_seq : function sep function_seq
                        | function function_seq
                        |'''
        self._parse_seq(p)

    def p_throws(self, p):
        '''throws : THROWS '(' field_seq ')'
                  |'''
        if len(p) == 5:
            p[0] = p[3]
        else:
            p[0] = deque()

    def p_function_type(self, p):
        '''function_type : field_type
                         | VOID'''
        if p[1] == 'void':
            p[0] = None
        else:
            p[0] = p[1]

    def p_field_seq(self, p):
        '''field_seq : field sep field_seq
                     | field field_seq
                     |'''
        self._parse_seq(p)

    def p_field(self, p):
        '''field : field_id field_req field_type IDENTIFIER annotations
                 | field_id field_req field_type IDENTIFIER '=' const_value \
                   annotations'''

        if len(p) == 8:
            default = p[6]
            annotations = p[7]
        else:
            default = None
            annotations = p[5]

        p[0] = ast.Field(
            id=p[1],
            name=p[4],
            field_type=p[3],
            requiredness=p[2],
            default=default,
            annotations=annotations,
            lineno=p.lineno(4),
        )

    def p_field_id(self, p):
        '''field_id : INTCONSTANT ':'
                    | '''
        if len(p) == 3:

            if p[1] == 0:
                # Prevent users from ever using field ID 0. It's reserved for
                # internal use only.
                raise ThriftParserError(
                    'Line %d: Field ID 0 is reserved for internal use.'
                    % p.lineno(1)
                )

            p[0] = p[1]
        else:
            p[0] = None

    def p_field_req(self, p):
        '''field_req : REQUIRED
                     | OPTIONAL
                     |'''
        if len(p) == 2:
            p[0] = p[1] == 'required'
        else:
            p[0] = None  # don't have a default

    def p_field_type(self, p):
        '''field_type : ref_type
                      | definition_type'''
        p[0] = p[1]

    def p_ref_type(self, p):
        '''ref_type : IDENTIFIER'''
        p[0] = ast.DefinedType(p[1], lineno=p.lineno(1))

    def p_base_type(self, p):  # noqa
        '''base_type : BOOL annotations
                     | BYTE annotations
                     | I16 annotations
                     | I32 annotations
                     | I64 annotations
                     | DOUBLE annotations
                     | STRING annotations
                     | BINARY annotations'''

        p[0] = ast.PrimitiveType(p[1], p[2])

    def p_container_type(self, p):
        '''container_type : map_type
                          | list_type
                          | set_type'''
        p[0] = p[1]

    def p_map_type(self, p):
        '''map_type : MAP '<' field_type ',' field_type '>' annotations'''
        p[0] = ast.MapType(key_type=p[3], value_type=p[5], annotations=p[7])

    def p_list_type(self, p):
        '''list_type : LIST '<' field_type '>' annotations'''
        p[0] = ast.ListType(value_type=p[3], annotations=p[5])

    def p_set_type(self, p):
        '''set_type : SET '<' field_type '>' annotations'''
        p[0] = ast.SetType(value_type=p[3], annotations=p[5])

    def p_definition_type(self, p):
        '''definition_type : base_type
                           | container_type'''
        p[0] = p[1]

    def p_annotations(self, p):
        '''annotations : '(' annotation_seq ')'
                       |'''
        if len(p) == 1:
            p[0] = []
        else:
            p[0] = list(p[2])

    def p_annotation_seq(self, p):
        '''annotation_seq : annotation sep annotation_seq
                          | annotation annotation_seq
                          |'''
        self._parse_seq(p)

    def p_annotation(self, p):
        '''annotation : IDENTIFIER '=' LITERAL '''
        p[0] = ast.Annotation(p[1], p[3], lineno=p.lineno(1))

    def _parse_seq(self, p):
        """Helper to parse sequence rules.

        Sequence rules are in the form::

            foo : foo_item sep foo
                | foo_item foo
                |

        This function builds a deque of the items in-order.

        If the number of tokens doesn't match, an exception is raised.
        """
        # This basically says:
        #
        # - When you reach the end of the list, construct and return an empty
        #   deque.
        # - Otherwise, prepend to start of what you got from the parser.
        #
        # So this ends up constructing an in-order list.
        if len(p) == 4:
            p[3].appendleft(p[1])
            p[0] = p[3]
        elif len(p) == 3:
            p[2].appendleft(p[1])
            p[0] = p[2]
        elif len(p) == 1:
            p[0] = deque()
        else:
            raise ThriftParserError(
                'Wrong number of tokens received for expression at line %d'
                % p.lineno(1)
            )


class Parser(ParserSpec):
    """Parser for Thrift IDL files."""

    def __init__(self, **kwargs):
        if kwargs.pop('silent', False):
            kwargs['errorlog'] = yacc.NullLogger()

        kwargs.setdefault('debug', False)
        kwargs.setdefault('write_tables', False)
        self._parser = yacc.yacc(module=self, **kwargs)
        self._lexer = Lexer()

    def parse(self, input, **kwargs):
        """Parse the given input.

        :param input:
            String containing the text to be parsed.
        :raises ThriftParserError:
            For parsing errors.
        """
        return self._parser.parse(input, lexer=self._lexer, **kwargs)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from thriftrw import spec
from thriftrw.spec.const import ConstValuePrimitive

__all__ = ['Generator']


class Generator(object):
    """Implements the Generate step of the compiler.

    During the Generate step, the system goes through the Thrift AST and
    collects all constants, and builds up type specs and classes with
    references to each other.
    """

    __slots__ = ('scope', 'type_mapper')

    def __init__(self, scope):
        """Initialize the generator.

        :param thriftrw.compile.scope.Scope scope:
            Scope maintaining the current compilation state.
        """
        self.scope = scope

    def process(self, definition):
        """Process the given definition from the AST.

        :param definition:
            A definition from the AST.
        """
        definition.apply(self)

    def visit_const(self, const):
        const_spec = spec.ConstSpec.compile(const)
        self.scope.add_const_spec(const_spec)

    def visit_typedef(self, typedef):
        self.scope.add_type_spec(
            typedef.name,
            spec.TypedefTypeSpec.compile(typedef),
            typedef.lineno,
        )

    def visit_enum(self, enum):
        enum_spec = spec.EnumTypeSpec.compile(enum)
        for key, value in enum_spec.items.items():
            self.scope.add_const_spec(
                spec.ConstSpec(
                    name=enum_spec.name + '.' + key,
                    value_spec=ConstValuePrimitive(value),
                    type_spec=spec.I32TypeSpec,
                    save=False,
                )
            )
        self.scope.add_type_spec(enum.name, enum_spec, enum.lineno)

    def visit_struct(self, struct):
        struct_spec = spec.StructTypeSpec.compile(struct)
        self.scope.add_type_spec(struct_spec.name, struct_spec, struct.lineno)

    def visit_union(self, union):
        union_spec = spec.UnionTypeSpec.compile(union)
        self.scope.add_type_spec(union_spec.name, union_spec, union.lineno)

    def visit_exc(self, exc):
        exc_spec = spec.ExceptionTypeSpec.compile(exc)
        self.scope.add_type_spec(exc_spec.name, exc_spec, exc.lineno)

    def visit_service(self, svc):
        service_spec = spec.ServiceSpec.compile(svc)
        self.scope.add_service_spec(service_spec)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from .reference import TypeReference
from .map import MapTypeSpec
from .set import SetTypeSpec
from .list import ListTypeSpec
from .primitive import (
    BoolTypeSpec,
    ByteTypeSpec,
    DoubleTypeSpec,
    I16TypeSpec,
    I32TypeSpec,
    I64TypeSpec,
    BinaryTypeSpec,
    TextTypeSpec,
)


__slots__ = ['type_spec_or_ref']


#: Mapping of Thrift primitive type names to corresponding type specs.
PRIMITIVE_TYPES = {
    'bool': BoolTypeSpec,
    'byte': ByteTypeSpec,
    'double': DoubleTypeSpec,
    'i16': I16TypeSpec,
    'i32': I32TypeSpec,
    'i64': I64TypeSpec,
    'string': TextTypeSpec,
    'binary': BinaryTypeSpec,
}


class TypeSpecMapper(object):
    """Maps AST types to type specifications.

    For references to defined types, return ``TypeReference`` objects instead.
    These will be replaced during the linking stage.
    """

    __slots__ = ()

    def get(self, typ):
        """Get the TypeSpec for the given AST type.

        If the type being referenced is a custom defined type, a TypeReference
        is returned instead.
        """
        return typ.apply(self)

    def visit_defined(self, typ):
        return TypeReference(typ.name, typ.lineno)

    def visit_primitive(self, typ):
        assert typ.name in PRIMITIVE_TYPES
        return PRIMITIVE_TYPES[typ.name]

    def visit_map(self, mtype):
        kspec = self.get(mtype.key_type)
        vspec = self.get(mtype.value_type)
        return MapTypeSpec(kspec, vspec)

    def visit_set(self, stype):
        vspec = self.get(stype.value_type)
        return SetTypeSpec(vspec)

    def visit_list(self, ltype):
        vspec = self.get(ltype.value_type)
        return ListTypeSpec(vspec)


#: Gets a TypeSpec for the given type if it's a native Thrift type or a
#: TypeReference if it's a custom defined type.
type_spec_or_ref = TypeSpecMapper().get



#!/usr/bin/env python

from setuptools import setup, find_packages

with open('README.rst') as f:
    long_description = f.read()

setup(
    name='thriftrw',
    version='0.3.1',
    description=(
        'A library to serialize and deserialize Thrift values.'
    ),
    long_description=long_description,
    author='Abhinav Gupta',
    author_email='abg@uber.com',
    url='https://github.com/uber/thriftrw-python',
    packages=find_packages(exclude=('tests', 'tests.*')),
    license='MIT',
    install_requires=['six', 'ply'],
    tests_require=['pytest', 'mock'],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ]
)



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from __future__ import absolute_import, unicode_literals, print_function

import sys
import inspect
import os.path

from .idl import Parser
from .compile import Compiler
from .protocol import BinaryProtocol


class Loader(object):
    """Loads and compiles Thrift files."""

    __slots__ = ('parser', 'compiler', 'compiled_modules')

    def __init__(self, protocol=None):
        """Initialize a loader.

        :param thriftrw.protocol.Protocol protocol:
            The protocol to use to serialize and deserialize types. Defaults
            to the binary protocol.
        """
        protocol = BinaryProtocol()

        self.parser = Parser()
        self.compiler = Compiler(protocol)

        # Mapping of absolute file path to compiled module. This is used to
        # cache the result of calling load() multiple times on the same file.
        self.compiled_modules = {}

    def loads(self, name, document):
        """Parse and compile the given Thrift document.

        :param str name:
            Name of the Thrift document.
        :param str document:
            The Thrift IDL as a string.
        """
        program = self.parser.parse(document)
        return self.compiler.compile(name, program)

    def load(self, path, name=None, force=False):
        """Load and compile the given Thrift file.

        If the file was already compiled before, a cached copy of the compiled
        module is returned.

        :param str path:
            Path to the ``.thrift`` file.
        :param str name:
            Name of the generated module. Defaults to the base name of the
            file.
        :param bool force:
            Whether to ignore the cache and load the file anew. Defaults to
            False.
        :returns:
            The compiled module.
        """
        path = os.path.abspath(path)
        if path in self.compiled_modules and not force:
            return self.compiled_modules[path]

        if name is None:
            # TODO do we care if the file extension is .thrift?
            name = os.path.splitext(os.path.basename(path))[0]
        with open(path, 'r') as f:
            document = f.read()
        module = self.loads(name, document)
        self.compiled_modules[path] = module
        return module


_DEFAULT_LOADER = Loader()

#: Parses and compiles the given Thrift file.
#:
#: Uses the binary protocol to serialize and deserialize values.
#:
#: For more advanced use, see :py:class:`thriftrw.loader.Loader`.
#:
#: :param str path:
#:     Path to the ``.thrift`` file.
#: :param str name:
#:     Name of the generated module. Defaults to the basename of the file.
#: :param bool force:
#:     Whether to ignore the cache and load the file anew. Defaults to False.
#: :raises thriftrw.compile.ThriftCompilerError:
#:     If there was an error compiling the Thrift file.
load = _DEFAULT_LOADER.load


def install(path, name=None):
    """Compiles a Thrift file and installs it as a submodule of the caller.

    Given a tree organized like so::

        foo/
            __init__.py
            bar.py
            my_service.thrift

    You would do,

    .. code-block:: python

        my_service = thriftrw.install('my_service.thrift')

    To install ``my_service`` as a submodule of the module from which you made
    the call. If the call was made in ``foo/bar.py``, the compiled
    Thrift file will be installed as ``foo.bar.my_service``. If the call was
    made in ``foo/__init__.py``, the compiled Thrift file will be installed as
    ``foo.my_service``. This allows other modules to import ``from`` the
    compiled module like so,

    .. code-block:: python

        from foo.my_service import MyService

    .. versionadded:: 0.2

    :param path:
        Path of the Thrift file. This may be an absolute path, or a path
        relative to the Python module making the call.
    :param str name:
        Name of the submodule. Defaults to the basename of the Thrift file.
    :returns:
        The compiled module
    """
    if name is None:
        name = os.path.splitext(os.path.basename(path))[0]

    callermod = inspect.getmodule(inspect.stack()[1][0])
    name = '%s.%s' % (callermod.__name__, name)

    if name in sys.modules:
        return sys.modules[name]

    if not os.path.isabs(path):
        callerfile = callermod.__file__
        path = os.path.normpath(
            os.path.join(os.path.dirname(callerfile), path)
        )

    sys.modules[name] = mod = load(path, name=name)
    return mod


__all__ = ['Loader', 'load', 'install']



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function

from collections import namedtuple

__all__ = [
    'Program',
    'Include',
    'Namespace',
    'Const',
    'Typedef',
    'Enum',
    'EnumItem',
    'Struct',
    'Union',
    'Exc',
    'Service',
    'Function',
    'Field',
    'PrimitiveType',
    'MapType',
    'SetType',
    'ListType',
    'DefinedType',
    'ConstValue',
    'ConstPrimitiveValue',
    'ConstReference',
    'ConstList',
    'ConstMap',
    'Annotation',
]


##############################################################################
# Program


class Program(namedtuple('Program', 'headers definitions')):
    """The top-level object representing the full Thrift IDL.

    The following attributes are available:

    .. py:attribute:: headers

        Collection of :py:class:`Include` and :py:class:`Namespace` objects.

    .. py:attribute:: definitions

        Collection of items of the following types.

        - :py:class:`Const`
        - :py:class:`Typedef`
        - :py:class:`Enum`
        - :py:class:`Struct`
        - :py:class:`Union`
        - :py:class:`Exc`
        - :py:class:`Service`
    """


##############################################################################
# Headers


class Include(namedtuple('Include', 'path lineno')):
    """A request to include the Thrift file at the given path.

    ::

        include "common.thrift"

    .. py:attribute:: path

        Path to the file to be included.
    """

    def apply(self, visitor):
        return visitor.visit_include(self)


class Namespace(namedtuple('Namespace', 'scope name lineno')):
    """Used to specify an alternative namespace for the given scope.

    ::

        namespace py my_service.generated

    .. py:attribute:: scope

        Scope to which this namespace rule applies. Can be a language
        specifier or '*' to apply to all languages.

    .. py:attribute:: name

        Namespace for the specified scope.
    """

    def apply(self, visitor):
        return visitor.visit_namespace(self)


##############################################################################
# Definitions


class Const(namedtuple('Const', 'name value_type value lineno')):
    """A constant defined in the Thrift IDL.

    ::

        const i32 DEFAULT_ID = 0;

    .. py:attribute:: name

        Name of the constant.

    .. py:attribute:: value_type

        Type of value held by the constant.

    .. py:attribute:: value

        Value specified for the constant.
    """

    def apply(self, visitor):
        return visitor.visit_const(self)


class Typedef(namedtuple('Typedef', 'name target_type annotations lineno')):
    """Typedefs define a new type which is an alias for another type.

    ::

        typedef string UUID

    .. py:attribute:: name

        Name of the new defined type.

    .. py:attribute:: target_type

        Type being aliased.

    .. py:attribute:: annotations

        Annotations for this type. See :py:class:`Annotation`.
    """

    def apply(self, visitor):
        return visitor.visit_typedef(self)


class Enum(namedtuple('Enum', 'name items annotations lineno')):
    """Enums define a new type which is a set of named integer values.

    ::

        enum Role {
            USER = 1,
            ADMIN,
        }

    .. py:attribute:: name

        Name of the enum type.

    .. py:attribute:: items

        Items defined in this Enum type. See :py:class:`EnumItem`.

    .. py:attribute:: annotations

        Annotations for this type. See :py:class:`Annotation`.
    """

    def apply(self, visitor):
        return visitor.visit_enum(self)


class EnumItem(namedtuple('EnumItem', 'name value annotations lineno')):
    """An item defined in an :py:class:`Enum` definition.

    .. py:attribute:: name

        Name of the item.

    .. py:attribute:: value

        Value specified for this item, if any. ``None`` othewise.

    .. py:attribute:: annotations

        Annotations for this item. See :py:class:`Annotation`.
    """


class Struct(namedtuple('Struct', 'name fields annotations lineno')):
    """A struct is a collection of named fields.

    .. py:attribute:: name

        Name of the struct.

    .. py:attribute:: fields

        Fields defined in the struct. See :py:class:`Field`.

    .. py:attribute:: annotations

        Annotations for this type. See :py:class:`Annotation`.
    """

    def apply(self, visitor):
        return visitor.visit_struct(self)


class Union(namedtuple('Union', 'name fields annotations lineno')):
    """A union is a sum of different types.

    .. py:attribute:: name

        Name of the union.

    .. py:attribute:: fields

        Fields defined in the union. See :py:class:`Field`.

    .. py:attribute:: annotations

        Annotations for this type. See :py:class:`Annotation`.
    """

    def apply(self, visitor):
        return visitor.visit_union(self)


class Exc(namedtuple('Exc', 'name fields annotations lineno')):
    """A Thrift exception definition.

    .. py:attribute:: name

        Name of the exception class.

    .. py:attribute:: fields

        Fields defined for the class. See :py:class:`Field`.

    .. py:attribute:: annotations

        Annotations for this type. See :py:class:`Annotation`.
    """

    def apply(self, visitor):
        return visitor.visit_exc(self)

# Can't use the name Exception because that will shadow the Exception class
# defined by Python.


class Service(
    namedtuple('Service', 'name functions parent annotations lineno')
):
    """A service definition.

    .. py:attribute:: name

        Name of the service.

    .. py:attribute:: functions

        Collection of function defined in the service. See
        :py:class:`Function`.

    .. py:attribute:: parent

        Name of the service that this service extends. ``None`` if this service
        doesn't have a parent service.

    .. py:attribute:: annotations

        Annotations for this service. See :py:class:`Annotation`.
    """

    def apply(self, visitor):
        return visitor.visit_service(self)


class Function(
    namedtuple(
        'Function',
        'name parameters return_type exceptions oneway annotations lineno'
    )
):
    """A function defined inside a service.

    .. py:attribute:: name

        Name of the function.

    .. py:attribute:: parameters

        Collection of parameters accepted by this method. See
        :py:class:`Field`.

    .. py:attribute:: return_type

        The type of value returned by this method.

    .. py:attribute:: exceptions

        Collection of exceptions raised by this method. See :py:class:`Field`.

    .. py:attribute:: oneway

        Whether this method is ``oneway`` or not.

    .. py:attribute:: annotations

        Annotations for this method. See :py:class:`Annotation`.
    """


class Field(
    namedtuple(
        'Field',
        'id name field_type requiredness default annotations lineno'
    )
):
    """A field defined inside a struct, union, exception, or parameter list.

    .. py:attribute:: id

        The numeric field identifier. `None` if not specified.

    .. py:attribute:: name

        Name of the field.

    .. py:attribute:: field_type

        Type of value held by the field.

    .. py:attribute:: requiredness

        ``True`` if this field was ``required``, ``False`` if ``optional``.
        ``None`` if required or optional was not specified.

    .. py:attribute:: default

        Default value of the field. ``None`` if not specified.

    .. py:attribute:: annotations

        Annotations for this field. See :py:class:`Annotation`.
    """

##############################################################################
# Types


class PrimitiveType(namedtuple('PrimitiveType', 'name annotations')):
    """Reference to primitive types.

    .. py:attribute:: name

        Name of the primitive type.

    .. py:attribute:: annotations

        Annotations for this type. See :py:class:`Annotation`.
    """

    def apply(self, visitor):
        return visitor.visit_primitive(self)


class MapType(namedtuple('MapType', 'key_type value_type annotations')):
    """A ``map<key, value>`` type.

    .. py:attribute:: key_type

        Type of the keys in the map.

    .. py:attribute:: value_type

        Type of the values in the map.

    .. py:attribute:: annotations

        Annotations for this type. See :py:class:`Annotation`.
    """

    def apply(self, visitor):
        return visitor.visit_map(self)


class SetType(namedtuple('SetType', 'value_type annotations')):
    """A ``set<item>`` type.

    .. py:attribute:: value_type

        Type of items in the set.

    .. py:attribute:: annotations

        Annotations for this type. See :py:class:`Annotation`.
    """

    def apply(self, visitor):
        return visitor.visit_set(self)


class ListType(namedtuple('ListType', 'value_type annotations')):
    """A ``list<item>`` type.

    .. py:attribute:: value_type

        Type of items in the list.

    .. py:attribute:: annotations

        Annotations for this type. See :py:class:`Annotation`.
    """

    def apply(self, visitor):
        return visitor.visit_list(self)


class DefinedType(namedtuple('DefinedType', 'name lineno')):
    """Reference to a type defined by the user.

    .. py:attribute:: name

        Name of the referenced type.
    """

    def apply(self, visitor):
        return visitor.visit_defined(self)


##############################################################################
# Constants

# TODO Move things that are sectioned like this into separate modules and
# explicitly specify their visitor interfaces.

class ConstValue(object):
    """Base class for constant value types."""

    def apply(self, visitor):
        raise NotImplementedError


class ConstPrimitiveValue(
    namedtuple('ConstPrimitiveValue', 'value lineno'), ConstValue
):
    """A complete constant value.

    .. py:attribute:: value

        Value held in this constant.
    """

    def apply(self, visitor):
        return visitor.visit_primitive(self)


class ConstReference(namedtuple('ConstReference', 'name lineno'), ConstValue):
    """Reference to another constant value or enum item.

    .. py:attribute:: name

        Name of the constant or enum item.
    """

    def apply(self, visitor):
        return visitor.visit_reference(self)


class ConstList(namedtuple('ConstList', 'values lineno'), ConstValue):
    """A list of constant values.

    .. py:attribute:: values

        Collection of ``ConstValue`` objects.
    """

    def apply(self, visitor):
        return visitor.visit_list(self)


class ConstMap(namedtuple('ConstList', 'pairs lineno'), ConstValue):
    """A map of constant values.

    .. py:attribute:: pairs

        Collection of pairs of ``ConstValue`` objects.
    """

    def apply(self, visitor):
        return visitor.visit_map(self)


##############################################################################
# Other


class Annotation(namedtuple('Annotation', 'name value lineno')):
    """Annotations for entities that can be annotated.

    They're usually in the form,::

        (foo = "bar", baz = "qux")

    For example,::

        struct User {
            1: string name (sensitive = "true");
        }

    .. py:attribute:: name

        Name of the annotation.

    .. py:attribute:: value

        Value specified for the annotation.
    """



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.
from __future__ import absolute_import, unicode_literals, print_function

import abc

from collections import namedtuple

from .ttype import TType

# TODO: This module should mostly be cython.

__all__ = [
    'Value',
    'BoolValue',
    'ByteValue',
    'DoubleValue',
    'I16Value',
    'I32Value',
    'I64Value',
    'BinaryValue',
    'FieldValue',
    'StructValue',
    'MapValue',
    'SetValue',
    'ListValue',
    'ValueVisitor',
]


class Value(object):
    """Base class for Value classes.

    Value classes define an intermediate representation of Thrift types as
    they are sent and received over the wire.
    """
    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def apply(self, visitor):
        """Apply the value to the given visitor.

        The appropriate `visit_*` method will be called and its result
        returned.

        :param ValueVisitor visitor:
            Visitor on the value.
        :returns:
            Value returned by the corresponding ``visit_*`` method.
        """
        pass


class BoolValue(namedtuple('BoolValue', 'value'), Value):
    """Wrapper for boolean values."""

    ttype_code = TType.BOOL

    def apply(self, visitor):
        return visitor.visit_bool(self.value)


class ByteValue(namedtuple('ByteValue', 'value'), Value):
    """Wrapper for byte values."""

    ttype_code = TType.BYTE

    def apply(self, visitor):
        return visitor.visit_byte(self.value)


class DoubleValue(namedtuple('DoubleValue', 'value'), Value):
    """Wrapper for double values."""

    ttype_code = TType.DOUBLE

    def apply(self, visitor):
        return visitor.visit_double(self.value)


class I16Value(namedtuple('I16Value', 'value'), Value):
    """Wrapper for 16-bit integer values."""

    ttype_code = TType.I16

    def apply(self, visitor):
        return visitor.visit_i16(self.value)


class I32Value(namedtuple('I32Value', 'value'), Value):
    """Wrapper for 32-bit integer values."""

    ttype_code = TType.I32

    def apply(self, visitor):
        return visitor.visit_i32(self.value)


class I64Value(namedtuple('I64Value', 'value'), Value):
    """Wrapper for 64-bit integer values."""

    ttype_code = TType.I64

    def apply(self, visitor):
        return visitor.visit_i64(self.value)


class BinaryValue(namedtuple('BinaryValue', 'value'), Value):
    """Wrapper for binary blobs.

    Note that Thrift does not differentiate between text and binary blobs over
    the wire. UTF-8 text should be encoded/decoded manually.
    """

    ttype_code = TType.BINARY

    def apply(self, visitor):
        return visitor.visit_binary(self.value)


class FieldValue(namedtuple('FieldValue', 'id ttype value'), Value):
    """A single field in a struct.

    .. py:attribute:: id

        Field identifier.

    .. py:attribute:: ttype

        :py:data:`~thriftrw.wire.TType` of the value held in this field.

    .. py:attribute:: value

        Value for this field.
    """


class StructValue(Value):
    """A struct value is a collection of fields of different types.

    .. py:attribute:: fields

        Collection of :py:class:`FieldValue` objects.
    """

    ttype_code = TType.STRUCT

    __slots__ = ('fields', '_index')

    def __init__(self, fields):
        self.fields = fields

        self._index = {}
        for field in fields:
            self._index[(field.id, field.ttype)] = field

    def apply(self, visitor):
        return visitor.visit_struct(self.fields)

    def get(self, field_id, field_ttype):
        """Returns the value at the given field ID and type.

        :param field_id:
            Numerical field identifier.
        :param field_ttype:
            Type of the value.
        :returns:
            Corresponding ``FieldValue`` or None.
        """
        field_value = self._index.get((field_id, field_ttype))
        if field_value is None:
            return None
        else:
            return field_value

    def __str__(self):
        return 'StructValue(%r)' % self.fields

    __repr__ = __str__

    def __eq__(self, other):
        return self.fields == other.fields


class MapValue(namedtuple('MapValue', 'key_ttype value_ttype pairs'), Value):
    """A mapping of two different kinds of values.

    This object may be treated as a map.

    .. py:attribute:: key_ttype

        Type of the keys stored in the map. This must be a value from TType.

    .. py:attribute:: value_ttype

        Type of the values stored in the map. This must be a value from TType.

    .. py:attribute:: pairs

        Collection of key-value tuples. Note that this is **not** a dict.
    """

    ttype_code = TType.MAP

    def apply(self, visitor):
        return visitor.visit_map(self.key_ttype, self.value_ttype, self.pairs)


class SetValue(namedtuple('SetValue', 'value_ttype values'), Value):
    """A collection of unique values of the same type.

    .. py:attribute:: value_ttype

        Type of values in the set. This must be a value from TType.

    .. py:attribute:: values

        Collection of the values.
    """

    ttype_code = TType.SET

    def apply(self, visitor):
        return visitor.visit_set(self.value_ttype, self.values)


class ListValue(namedtuple('ListValue', 'value_ttype values'), Value):
    """A collection of values.

    .. py:attribute:: value_ttype

        Type of values in the list. This must be a value from TType.

    .. py:attribute:: values

        Collection of the values.
    """

    ttype_code = TType.LIST

    def apply(self, visitor):
        return visitor.visit_list(self.value_ttype, self.values)


class ValueVisitor(object):
    """Visitor on different value types.

    The ``visit_*`` functions are not given the ``*Value`` objects but the
    actual values contained in them.

    The idea is that when something needs to take an arbitrary Thrift value
    and act on all cases, it extends this type. Each Thrift value knows which
    function to call on the visitor based on the value type. The intention
    here is to avoid ``isinstance`` checks.
    """

    __metaclass__ = abc.ABCMeta

    @abc.abstractmethod
    def visit_bool(self, value):
        """Visits boolean values.

        :param bool value:
            True or False
        """
        pass

    @abc.abstractmethod
    def visit_byte(self, value):
        """Visits 8-bit integers.

        :param int value:
            8-bit integer
        """
        pass

    @abc.abstractmethod
    def visit_double(self, value):
        """Visits double values.

        :param float value:
            Floating point number
        """
        pass

    @abc.abstractmethod
    def visit_i16(self, value):
        """Visits 16-bit integers.

        :param int value:
            16-bit integer
        """
        pass

    @abc.abstractmethod
    def visit_i32(self, value):
        """Visits 32-bit integers.

        :param int value:
            32-bit integer
        """
        pass

    @abc.abstractmethod
    def visit_i64(self, value):
        """Visits 64-bit integers.

        :param int value:
            64-bit integer
        """
        pass

    @abc.abstractmethod
    def visit_binary(self, value):
        """Visits binary blobs.

        :param bytes value:
            Binary blob
        """
        pass

    @abc.abstractmethod
    def visit_struct(self, fields):
        """Visits structs.

        :param fields:
            Collection of :py:class:`FieldValue` objects.
        """
        pass

    @abc.abstractmethod
    def visit_map(self, key_ttype, value_ttype, pairs):
        """Visits maps.

        :param thriftrw.wire.TType key_ttype:
            TType of the keys in the map.
        :param thriftrw.wire.TType value_ttype:
            TType of the values in the map.
        :param pairs:
            Collection of key-value pairs.
        """
        pass

    @abc.abstractmethod
    def visit_set(self, value_ttype, values):
        """Visits sets.

        :param thriftrw.wire.TType value_ttype:
            TType of the items in the set.
        :param values:
            Collection of values in the set.
        """
        pass

    @abc.abstractmethod
    def visit_list(self, value_ttype, values):
        """Visits lists.

        :param thriftrw.wire.TType value_ttype:
            TType of the items in the set.
        :param values:
            Collection of values in the list.
        """
        pass



# Copyright (c) 2015 Uber Technologies, Inc.
#
# Permission is hereby granted, free of charge, to any person obtaining a copy
# of this software and associated documentation files (the "Software"), to deal
# in the Software without restriction, including without limitation the rights
# to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
# copies of the Software, and to permit persons to whom the Software is
# furnished to do so, subject to the following conditions:
#
# The above copyright notice and this permission notice shall be included in
# all copies or substantial portions of the Software.
#
# THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
# IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
# FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
# AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
# LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
# OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
# THE SOFTWARE.

from __future__ import absolute_import, unicode_literals, print_function


__all__ = ['TypeSpecLinker', 'ServiceSpecLinker']


class TypeSpecLinker(object):
    """Links together references in TypeSpecs."""

    __slots__ = ('scope',)

    def __init__(self, scope):
        self.scope = scope

    def link(self):
        """Resolve and link all types in the scope."""

        type_specs = {}
        types = []

        for name, type_spec in self.scope.type_specs.items():
            type_spec = type_spec.link(self.scope)
            type_specs[name] = type_spec

            if type_spec.surface is not None:
                self.scope.add_surface(name, type_spec.surface)
                types.append(type_spec.surface)

        self.scope.type_specs = type_specs
        self.scope.add_surface('types', tuple(types))


class ServiceSpecLinker(object):
    """Links references to type in service specs."""

    __slots__ = ('scope',)

    def __init__(self, scope):
        self.scope = scope

    def link(self):
        service_specs = {}
        services = []

        for name, service_spec in self.scope.service_specs.items():
            service_spec = service_spec.link(self.scope)
            service_specs[name] = service_spec

            if service_spec.surface is not None:
                self.scope.add_surface(
                    service_spec.name, service_spec.surface
                )
                services.append(service_spec.surface)

        self.scope.service_specs = service_specs
        self.scope.add_surface('services', tuple(services))


class ConstSpecLinker(object):
    """Links references to top-level constants."""

    __slots__ = ('scope',)

    def __init__(self, scope):
        self.scope = scope

    def link(self):
        const_specs = {}
        constants = {}

        for name, const_spec in self.scope.const_specs.items():
            const_spec = const_spec.link(self.scope)
            const_specs[name] = const_spec

            if const_spec.surface is not None and const_spec.save:
                self.scope.add_surface(
                    const_spec.name, const_spec.surface
                )
                constants[const_spec.name] = const_spec.surface

        self.scope.const_specs = const_specs
        self.scope.add_surface('constants', constants)
